// Set Of Integer
// Author: Michal Kokorceny
// E-mail: michal@74.cz
// Web:    http://74.cz

{$A+,B-,C+,G+,H+,I+,J+,M-,O+,P+,Q-,R-,U-,V-,W-,Z-}

unit mksetint;

interface

uses
  classes;

type
  tsetofbyte=set of byte;

  tarrayofinteger=array of integer;

  tbuffer=array of record
    node:integer;
    bits:tsetofbyte;
  end;

  tsetofinteger=class
  private
    fbuffer:tbuffer;
    flength:integer;
    ffindindex:integer;
    ffindmask:byte;
    function flocate(i:integer;out index,node:integer;out mask:byte):boolean;
    procedure ftrimlength;
  public
    procedure clear;
    procedure assign(setint:tsetofinteger);
    procedure include(i:integer); overload;
    procedure include(a:array of integer); overload;
    procedure exclude(i:integer); overload;
    procedure exclude(a:array of integer); overload;
    procedure invert(i:integer); overload;
    procedure invert(a:array of integer); overload;
    procedure intersection(setint:tsetofinteger);
    procedure union(setint:tsetofinteger);
    procedure difference(setint:tsetofinteger);
    function subset(setint:tsetofinteger):boolean;
    function notsubset(setint:tsetofinteger):boolean;
    function superset(setint:tsetofinteger):boolean;
    function notsuperset(setint:tsetofinteger):boolean;
    function equality(setint:tsetofinteger):boolean;
    function inequality(setint:tsetofinteger):boolean;
    function within(i:integer):boolean;
    function count:integer;
    function isempty:boolean;
    function findfirst(out i:integer):boolean;
    function findnext(out i:integer):boolean;
    function exportasarray:tarrayofinteger;
    procedure loadfromstream(stream:tstream);
    procedure savetostream(stream:tstream);
    procedure loadfromfile(const filename:string);
    procedure savetofile(const filename:string);
    constructor create;
    destructor destroy; override;
  end;

implementation

uses
  sysutils;

const
  delta1=32;
  delta2=128;
  numberofbits:array[0..15] of integer=(0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4);  // pocet bitu v 4-bitovem cisle
  signature:integer=$31544E49;

type
  tset32=set of 0..31;                   
  tarray32=packed array[0..31] of byte;  // pretypovani mnoziny na pole 32 bytu
  ESetOfInteger=class(exception);        // vyjimka pro set of integer

resourcestring
  e_crc='CRC error';

// najit prvek
// i     - prvek
// index - index
// node  - uzel
// mask  - maska
function tsetofinteger.flocate(i:integer;out index,node:integer;out mask:byte):boolean;
var
  x,x1,x2:integer;
begin
  if i<0 then begin
    node:=((i+1) div 256)-1;
    mask:=255-(-i-1) mod 256;
  end
  else begin
    node:=i div 256;
    mask:=i mod 256;
  end;
  if flength>0 then begin
    x1:=0;
    x2:=flength-1;
    repeat
      if node<=fbuffer[x1].node then begin
        index:=x1;
        break;
      end
      else if node=fbuffer[x2].node then begin
        index:=x2;
        break;
      end
      else if node>fbuffer[x2].node then begin
        index:=x2+1;
        break;
      end;
      x:=(x1+x2) div 2;
      if node>fbuffer[x].node then x1:=x+1 else if node<=fbuffer[x].node then x2:=x-1;
    until false;
  end
  else index:=0;
  result:=(index<flength) and (node=fbuffer[index].node);
end;

// upravit delku bufferu
procedure tsetofinteger.ftrimlength;
var
  x:integer;
begin
  try
    x:=length(fbuffer)-flength;
    if (x<0) or (x>=delta2) then begin
      x:=flength mod delta1;
      if x<>0 then x:=delta1-x;
      setlength(fbuffer,flength+x);
    end;
  except
    if length(fbuffer)<flength then clear;
    raise;
  end;
end;

// vyprazdnit mnozinu
procedure tsetofinteger.clear;
begin
  flength:=0;
  fbuffer:=nil;
  ffindindex:=-1;
end;

// zkopirovat obsah mnoziny
// setint - zdrojova mnozina
procedure tsetofinteger.assign(setint:tsetofinteger);
begin
  try
    flength:=setint.flength;
    setlength(fbuffer,length(setint.fbuffer));
    move(setint.fbuffer[0],fbuffer[0],flength*sizeof(fbuffer[0]));
  except
    clear;
    raise;
  end;
end;

// vlozit prvek do mnoziny
// i - prvek
procedure tsetofinteger.include(i:integer);
var
  index,node:integer;
  mask:byte;
begin
  if not(flocate(i,index,node,mask)) then begin
    if flength=length(fbuffer) then setlength(fbuffer,length(fbuffer)+delta1);
    move(fbuffer[index],fbuffer[index+1],(flength-index)*sizeof(fbuffer[index]));
    inc(flength);
    fbuffer[index].node:=node;
    fbuffer[index].bits:=[mask];
  end
  else system.include(fbuffer[index].bits,mask);
end;

// vlozit pole integeru do mnoziny
// a - pole integeru
procedure tsetofinteger.include(a:array of integer);
var
  i:integer;
begin
  for i:=low(a) to high(a) do include(a[i]);
end;

// odstranit prvek z mnoziny
// i - prvek
procedure tsetofinteger.exclude(i:integer);
var
  index,node:integer;
  mask:byte;
begin
  if flocate(i,index,node,mask) then if fbuffer[index].bits=[mask] then begin
    move(fbuffer[index+1],fbuffer[index],(flength-index-1)*sizeof(fbuffer[index]));
    dec(flength);
    if length(fbuffer)-flength>=delta2 then setlength(fbuffer,flength);
  end
  else system.exclude(fbuffer[index].bits,mask);
end;

// odstranit pole integeru z mnoziny
// a - pole integeru
procedure tsetofinteger.exclude(a:array of integer);
var
  i:integer;
begin
  for i:=low(a) to high(a) do exclude(a[i]);
end;

// invertovat prvek v mnozine
// i - prvek
procedure tsetofinteger.invert(i:integer);
begin
  if within(i) then exclude(i) else include(i);
end;

// invertovat pole integeru v mnozine
// a - pole integeru
procedure tsetofinteger.invert(a:array of integer);
var
  i:integer;
begin
  for i:=low(a) to high(a) do invert(a[i]);
end;

// prunik dvou mnozin [SELF:=SELF*SETINT]
// setint - druha mnozina
procedure tsetofinteger.intersection(setint:tsetofinteger);
var
  i,s,d:integer;
begin
  s:=0;
  d:=0;
  for i:=0 to flength-1 do begin
    while (s<setint.flength) and (setint.fbuffer[s].node<fbuffer[i].node) do inc(s);
    if (s<setint.flength) and (setint.fbuffer[s].node=fbuffer[i].node) then begin
      fbuffer[d].bits:=fbuffer[i].bits*setint.fbuffer[s].bits;
      if fbuffer[d].bits<>[] then begin
        fbuffer[d].node:=fbuffer[i].node;
        inc(d);
      end;
    end;
  end;
  flength:=d;
  ftrimlength;
end;

// sjednoceni dvou mnozin [SELF:=SELF+SETINT]
// setint - druha mnozina
procedure tsetofinteger.union(setint:tsetofinteger);
var
  i,s,d:integer;
  dbuffer:tbuffer;
begin
  setlength(dbuffer,flength+setint.flength);
  i:=0;
  s:=0;
  d:=0;
  while (i<flength) or (s<setint.flength) do
   if (s=setint.flength) or ((i<flength) and (fbuffer[i].node<setint.fbuffer[s].node)) then begin
    dbuffer[d]:=fbuffer[i];
    inc(d);
    inc(i);
  end
  else if (i=flength) or ((s<setint.flength) and (fbuffer[i].node>setint.fbuffer[s].node)) then begin
    dbuffer[d]:=setint.fbuffer[s];
    inc(d);
    inc(s);
  end
  else begin
    dbuffer[d].node:=fbuffer[i].node;
    dbuffer[d].bits:=fbuffer[i].bits+setint.fbuffer[s].bits;
    inc(d);
    inc(i);
    inc(s);
  end;
  fbuffer:=dbuffer;
  dbuffer:=nil;
  flength:=d;
  ftrimlength;
end;
                      
// rozdil dvou mnozin [SELF:=SELF-SETINT]
// setint - druha mnozina
procedure tsetofinteger.difference(setint:tsetofinteger);
var
  i,s,d:integer;
begin
  s:=0;
  d:=0;
  for i:=0 to flength-1 do begin
    while (s<setint.flength) and (setint.fbuffer[s].node<fbuffer[i].node) do inc(s);
    if (s<setint.flength) and (setint.fbuffer[s].node=fbuffer[i].node) then begin
      fbuffer[d].bits:=fbuffer[i].bits-setint.fbuffer[s].bits;
      if fbuffer[d].bits<>[] then begin
        fbuffer[d].node:=fbuffer[i].node;
        inc(d);
      end;
    end
    else begin
      fbuffer[d]:=fbuffer[i];
      inc(d);
    end;
  end;
  flength:=d;
  ftrimlength;
end;

// podmnozina [SETINT<=SELF]
// setint - druha mnozina
function tsetofinteger.subset(setint:tsetofinteger):boolean;
var
  i,s:integer;
begin
  i:=0;
  s:=0;
  result:=setint.flength>=flength;
  while result and (i<flength) do begin
    while (s<setint.flength) and (setint.fbuffer[s].node<fbuffer[i].node) do inc(s);
    result:=(s<setint.flength) and
            (fbuffer[i].node=setint.fbuffer[s].node) and
            (fbuffer[i].bits<=setint.fbuffer[s].bits);
    inc(i);
  end;
end;

// neni podmnozina [SETINT>SELF]
// setint - druha mnozina
function tsetofinteger.notsubset(setint:tsetofinteger):boolean;
begin
  result:=not(subset(setint));
end;

// nadmnozina [SETINT>=SELF]
// setint - druha mnozina
function tsetofinteger.superset(setint:tsetofinteger):boolean;
var
  i,s:integer;
begin
  i:=0;
  s:=0;
  result:=flength>=setint.flength;
  while result and (s<setint.flength) do begin
    while (i<flength) and (fbuffer[i].node<setint.fbuffer[s].node) do inc(i);
    result:=(i<flength) and
            (setint.fbuffer[s].node=fbuffer[i].node) and
            (setint.fbuffer[s].bits<=fbuffer[i].bits);
    inc(s);
  end;
end;

// neni nadmnozina [SETINT<SELF]
// setint - druha mnozina
function tsetofinteger.notsuperset(setint:tsetofinteger):boolean;
begin
  result:=not(superset(setint));
end;

// rovnost dvou mnozin
// setint - druha mnozina
// result - true  - mnoziny jsou stejne
//          false - mnoziny nejsou stejne
function tsetofinteger.equality(setint:tsetofinteger):boolean;
var
  i:integer;
begin
  i:=0;
  result:=flength=setint.flength;
  while result and (i<flength) and
        (fbuffer[i].node=setint.fbuffer[i].node) and
        (fbuffer[i].bits=setint.fbuffer[i].bits) do inc(i);
end;

// nerovnost dvou mnozin
// setint - druha mnozina
// result - true  - mnoziny nejsou stejne
//          false - mnoziny jsou stejne
function tsetofinteger.inequality(setint:tsetofinteger):boolean;
begin
  result:=not(equality(setint));
end;

// prvek je/neni v mnozine
// i      - prvek
// result - je/neni v mnozine
function tsetofinteger.within(i:integer):boolean;
var
  index,node:integer;
  mask:byte;
begin
  result:=flocate(i,index,node,mask) and (mask in fbuffer[index].bits);
end;

// pocet prvku v mnozine
// result - pocet prvku
function tsetofinteger.count:integer;
var
  i,j:integer;
begin
  result:=0;
  for i:=0 to flength-1 do for j:=0 to 31 do begin
    inc(result,numberofbits[(tarray32(fbuffer[i].bits)[j] and 15)]);
    inc(result,numberofbits[(tarray32(fbuffer[i].bits)[j] shr 4)]);
  end;
end;

// testovat jestli je mnozina prazdna
// result - mnozina je/neni prazdna
function tsetofinteger.isempty:boolean;
begin
  result:=flength=0;
end;

// najit prvni prvek v mnozine
// i      - prvek
// result - nalezeno/nenalezeno
function tsetofinteger.findfirst(out i:integer):boolean;
begin
  ffindindex:=0;
  ffindmask:=0;
  result:=findnext(i);
end;

// najit dalsi prvek v mnozine
// i      - prvek
// result - nalezeno/nenalezeno
function tsetofinteger.findnext(out i:integer):boolean;

procedure next;
begin
  if ffindmask=255 then begin
    inc(ffindindex);
    ffindmask:=0;
  end
  else inc(ffindmask);
end;

begin
  result:=false;                  
  if ffindindex<>-1 then begin
    while (ffindindex<flength) and not(ffindmask in fbuffer[ffindindex].bits) do next;
    if ffindindex<flength then with fbuffer[ffindindex] do begin
      if node<0
       then i:=(node+1)*256+ffindmask-256
       else i:=node*256+ffindmask;
      result:=true;
    end;
    next;
    if ffindindex>=flength then ffindindex:=-1;
  end;
end;

// exportovat mnozinu jako pole
// result - vystupni pole
function tsetofinteger.exportasarray:tarrayofinteger;
var
  i,x:integer;
  res:boolean;
begin
  setlength(result,count);
  x:=0;
  res:=findfirst(i);
  while res do begin
    result[x]:=i;
    inc(x);
    res:=findnext(i);
  end;
end;

// nacist mnozinu ze streamu
// stream - zdrojovy stream
procedure tsetofinteger.loadfromstream(stream:tstream);
var
  i,j:integer;
  crc,crc2:cardinal;
  relevant:tset32;
begin
  try
    stream.read(i,sizeof(i));
    if i=signature then begin  // novy format (s kompresi)
      crc:=0;
      stream.read(flength,sizeof(flength));
      ftrimlength;
      for i:=0 to flength-1 do with fbuffer[i] do begin
        stream.read(node,sizeof(node));
        if (i>0) and (node<=fbuffer[i-1].node) then raise esetofinteger.create(e_crc);
        stream.read(relevant,sizeof(relevant));
        for j:=0 to 31 do if j in relevant then begin
          stream.read(tarray32(bits)[j],sizeof(tarray32(bits)[j]));
          inc(crc,tarray32(bits)[j]);
        end
        else tarray32(bits)[j]:=0;
      end;
      stream.read(crc2,sizeof(crc2));
      if crc<>crc2 then raise esetofinteger.create(e_crc);
    end
    else begin                 // stary format (bez komprese)
      flength:=i;
      ftrimlength;
      stream.read(fbuffer[0],sizeof(fbuffer[0])*flength);
    end
  except
    clear;
    raise;
  end;
end;

// ulozit mnozinu do streamu
// stream - cilovy stream
procedure tsetofinteger.savetostream(stream:tstream);
var
  i,j:integer;
  crc:cardinal;
  relevant:tset32;
begin
  crc:=0;
  stream.write(signature,sizeof(signature));
  stream.write(flength,sizeof(flength));
  for i:=0 to flength-1 do with fbuffer[i] do begin
    stream.write(node,sizeof(node));
    relevant:=[];
    for j:=0 to 31 do if tarray32(bits)[j]<>0 then system.include(relevant,j);
    stream.write(relevant,sizeof(relevant));
    for j:=0 to 31 do if j in relevant then begin
      stream.write(tarray32(bits)[j],sizeof(tarray32(bits)[j]));
      inc(crc,tarray32(bits)[j]);
    end;
  end;
  stream.write(crc,sizeof(crc));
end;

// nacist mnozinu ze souboru
// filename - nazev souboru
procedure tsetofinteger.loadfromfile(const filename:string);
var
  stream:tfilestream;
begin
  stream:=tfilestream.create(filename,fmopenread);
  try
    loadfromstream(stream);
  finally
    stream.free;
  end;
end;

// ulozit mnozinu do souboru
// filename - nazev souboru
procedure tsetofinteger.savetofile(const filename:string);
var
  stream:tfilestream;
begin
  stream:=tfilestream.create(filename,fmcreate);
  try
    savetostream(stream);
  finally
    stream.free;
  end;
end;

// vytvorit mnozinu
constructor tsetofinteger.create;
begin
  clear;
end;

// uvolnit mnozinu
destructor tsetofinteger.destroy;
begin
  clear;
  inherited;
end;

end.
